@@ -1,11 +1,12 @@
 from abc import ABC, abstractmethod, abstractproperty
-import git
+import git, gitdb
 from os.path import basename
 from bossman.errors import BossmanError
 from bossman.logging import get_class_logger
 from datetime import datetime
 import gitdb.exc
 import yaml
+import threading
 
 def true(*args, **kwargs):
   return True
@@ -103,7 +104,8 @@ class DeleteChange(Change):
     return self.diff.a_blob.data_stream.read()
 
 class Notes:
-  def __init__(self, commit: git.Commit, ns: str = None):
+  def __init__(self, repo, commit: git.Commit, ns: str = None):
+    self.repo = repo
     self.commit = commit
     self.ns = ns.strip("/")
 
@@ -117,30 +119,33 @@ class Notes:
     return notes.get(k, default)
 
   def read(self) -> dict:
-    try:
-      args = []
-      if self.ns is not None:
-        args.extend(("--ref", self.ns))
-      args.extend(("show", self.commit.hexsha))
-      notes = self.commit.repo.git.notes(*args)
-      return yaml.safe_load(notes if notes else '{}')
-    except git.GitCommandError:
-      return {}
+    with self.repo._lock:
+      try:
+        args = []
+        if self.ns is not None:
+          args.extend(("--ref", self.ns))
+        args.extend(("show", self.commit.hexsha))
+        notes = self.commit.repo.git.notes(*args)
+        return yaml.safe_load(notes if notes else '{}')
+      except git.GitCommandError:
+        return {}
 
   def write(self, notes):
     try:
-      args = []
-      if self.ns is not None:
-        args.extend(("--ref", self.ns))
-      args.extend(("add", "-f", "-m", yaml.safe_dump(notes), self.commit.hexsha))
-      self.commit.repo.git.notes(*args)
+      with self.repo._lock:
+        args = []
+        if self.ns is not None:
+          args.extend(("--ref", self.ns))
+        args.extend(("add", "-f", "-m", yaml.safe_dump(notes), self.commit.hexsha))
+        self.commit.repo.git.notes(*args)
     except git.GitCommandError as err:
       raise BossmanError(err)
 
 
 
 class Revision:
-  def __init__(self, commit: git.Commit, diffs: git.Diff):
+  def __init__(self, repo, commit: git.Commit, diffs: git.Diff):
+    self.repo = repo
     self.commit = commit
     self.changes = dict(
       (diff.b_path or diff.a_path, Change.from_diff(diff))
@@ -150,87 +155,103 @@ class Revision:
 
   @property
   def id(self):
-    return short_rev(self.commit.hexsha)
+    with self.repo._lock:
+      return short_rev(self.commit.hexsha)
 
   @property
   def parent_id(self):
     try:
-      return short_rev(self.commit.parents[0].hexsha)
+      with self.repo._lock:
+        return short_rev(self.commit.parents[0].hexsha)
     except IndexError:
       return None
 
   @property
   def branches(self):
-    cmd = git.cmd.Git(self.commit.repo.working_tree_dir)
-    branches = cmd.branch(contains=self.id, format="%(refname:short)")
-    return branches.splitlines()
+    with self.repo._lock:
+      return self.repo.get_branches_containing(self.id)
 
   @property
   def date(self) -> datetime:
-    return self.commit.committed_datetime
+    with self.repo._lock:
+      return self.commit.committed_datetime
 
   @property
   def message(self) -> str:
-    return self.commit.message
+    with self.repo._lock:
+      return self.commit.message
 
   @property
   def short_message(self) -> str:
-    return self.message.split("\n").pop(0)
+    with self.repo._lock:
+      return self.message.split("\n").pop(0)
 
   @property
   def author_name(self) -> str:
-    return self.commit.author.name
+    with self.repo._lock:
+      return self.commit.author.name
 
   @property
   def author_email(self) -> str:
-    return self.commit.author.email
+    with self.repo._lock:
+      return self.commit.author.email
 
   @property
   def committer_name(self) -> str:
-    return self.commit.committer.name
+    with self.repo._lock:
+      return self.commit.committer.name
 
   @property
   def committer_email(self) -> str:
-    return self.commit.committer.email
+    with self.repo._lock:
+      return self.commit.committer.email
 
   @property
   def affected_paths(self) -> list:
-    return self.changes.keys()
+    with self.repo._lock:
+      return self.changes.keys()
 
   def show_path(self, path) -> dict:
-    contents = self.show_paths([path])
-    return contents.get(path, None)
+    with self.repo._lock:
+      contents = self.show_paths([path])
+      return contents.get(path, None)
 
   def show_paths(self, paths: list) -> dict:
-    contents = dict()
-    def get_blob_contents(blob):
-      if blob.path in paths:
-        contents[blob.path] = blob.data_stream.read()
-    visitor = TreeVisitor(get_blob_contents)
-    visitor(self.commit.tree)
-    return contents
+    with self.repo._lock:
+      contents = dict()
+      def get_blob_contents(blob):
+        if blob.path in paths:
+          contents[blob.path] = blob.data_stream.read()
+      visitor = TreeVisitor(get_blob_contents)
+      visitor(self.commit.tree)
+      return contents
 
   def get_changes(self, paths: str) -> Change:
-    return list(change for (path, change) in self.changes.items() if path in paths)
+    with self.repo._lock:
+      return list(change for (path, change) in self.changes.items() if path in paths)
 
   def get_change(self, path: str) -> Change:
-    return self.changes.get(path, None)
+    with self.repo._lock:
+      return self.changes.get(path, None)
 
   def get_notes(self, ns: str = None) -> Notes:
-    return Notes(self.commit, ns)
+    with self.repo._lock:
+      return Notes(self.repo, self.commit, ns)
 
   def __str__(self):
-    s = "[{}] {} {} | {}".format(self.id, self.short_message, self.author_name, self.date)
-    if len(self.changes):
-      s += "\n\n  "
-      s += "\n  ".join(str(change) for change in self.changes.values())
-      s += "\n"
-    return s
+    with self.repo._lock:
+      s = "[{}] {} {} | {}".format(self.id, self.short_message, self.author_name, self.date)
+      if len(self.changes):
+        s += "\n\n  "
+        s += "\n  ".join(str(change) for change in self.changes.values())
+        s += "\n"
+      return s
 
   def __rich_console__(self, console, options):
-    yield "[b]\[{}][/b] {} | {} {}".format(self.id, self.short_message, self.author_name, self.date)
-    # for change in self.changes.values():
-    #   yield change
+    with self.repo._lock:
+      yield "[b]\[{}][/b] {} | {} {}".format(self.id, self.short_message, self.author_name, self.date)
+      # for change in self.changes.values():
+      #   yield change
 
 class RevisionDetails:
   """
@@ -248,23 +269,79 @@ class RevisionDetails:
   def __str__(self):
     return "{} ({})".format(self.id, self.details)
 
+def synchronized(method):
+  def wrapper(self, *arg, **kws):
+    with self._lock:
+      return method(self, *arg, **kws)
+  return wrapper
+
 class Repo:
+
   def __init__(self, root):
     self.logger = get_class_logger(self)
-    self._repo = git.Repo(root)
+    try:
+      self._repo = git.Repo(root)
+    except git.InvalidGitRepositoryError:
+      raise RepoError("Not a git repository: {}.".format(root))
+    self._lock = threading.RLock()
 
   def config_writer(self):
     return self._repo.config_writer()
 
+  def config_reader(self):
+    return self._repo.config_reader()
+
+  @synchronized
+  def fetch_notes(self, ns: str = "*"):
+    conf = self.config_reader()
+    ref = "refs/notes/" + ns
+    for section in conf.sections():
+      if section.startswith("remote"):
+        remote = section.split(" ").pop().strip('"')
+        self._repo.git.fetch(remote, "+{}:{}".format(ref, ref))
+
+  @synchronized
+  def push_notes(self, ns: str = "*"):
+    conf = self.config_reader()
+    ref = "refs/notes/" + ns
+    for section in conf.sections():
+      if section.startswith("remote"):
+        remote = section.split(" ").pop().strip('"')
+        self._repo.git.push(remote, "+{}:{}".format(ref, ref))
+
+  @synchronized
   def rev_parse(self, rev: str) -> str:
     # We can't use Repo.rev_parse since it doesn't support
     # short hexshas or indirect refs such as tag names or branches.
     try:
       hexsha = self._repo.git.rev_parse(rev)
       return self._repo.rev_parse(hexsha)
-    except git.GitCommandError:
+    except (git.GitCommandError, gitdb.exc.BadName):
       raise BossmanError("failed to resolve revision {}, please make sure it is a valid commit/tag name".format(rev))
 
+  @synchronized
+  def rev_exists(self, rev: str) -> str:
+    try:
+      return isinstance(self.rev_parse(rev), git.Commit)
+    except BossmanError:
+      return False
+
+  @synchronized
+  def rev_is_reachable(self, rev: str, from_rev: str = "HEAD") -> str:
+    """
+    Returns True if {rev} is an ancestor of {from_rev}.
+    """
+    try:
+      rev = self.rev_parse(rev)
+      for parent_rev in self._repo.iter_commits(from_rev):
+        if parent_rev.hexsha == rev.hexsha:
+          return True
+      else:
+        return False
+    except BossmanError:
+      return False
+
+  @synchronized
   def get_paths(self, rev: str = "HEAD", predicate = true) -> list:
     """
     Lists the paths versioned for revision {rev}, optionally filtered
@@ -280,6 +357,7 @@ class Repo:
     visitor(commit.tree)
     return paths
 
+  @synchronized
   def get_head(self):
     """
     Returns the HEAD revision, or None if this is the first commit.
@@ -290,25 +368,27 @@ class Repo:
       if commit.parents:
         prev = commit.parents[0]
       diffs = commit.diff(prev, R=True)
-      return Revision(commit, diffs)
+      return Revision(self, commit, diffs)
     except StopIteration:
       return None
 
+  @synchronized
   def get_last_revision(self, paths: list, rev: str = "HEAD") -> Revision:
     """
     Returns the last revision in {rev}'s ancestry to have affected {paths}.
     """
     try:
-      rev = self.rev_parse(rev)
+      rev = self.rev_parse(rev) if rev is not None else "HEAD"
       commit = next(self._repo.iter_commits(rev, paths=paths))
       prev = git.NULL_TREE
       if commit.parents:
         prev = commit.parents[0]
       diffs = commit.diff(prev, paths=paths, R=True)
-      return Revision(commit, diffs)
+      return Revision(self, commit, diffs)
     except StopIteration:
       return None
 
+  @synchronized
   def get_revision(self, rev: str, paths: list = None):
     """
     Returns the revision {rev}, with diffs for {paths}.
@@ -319,11 +399,40 @@ class Repo:
     if commit.parents:
       prev = commit.parents[0]
       diffs = commit.diff(prev, paths=paths, R=True) # R=True -> reverse
-    return Revision(commit, diffs)
+    return Revision(self, commit, diffs)
 
+  @synchronized
+  def get_current_user_email(self):
+    """
+    Returns the result of `git config user.email`
+    """
+    return self._repo.git.config("user.email")
+
+  @synchronized
+  def get_current_user_name(self):
+    """
+    Returns the result of `git config user.name`
+    """
+    return self._repo.git.config("user.name")
+
+
+  @synchronized
   def get_current_branch(self):
     return self._repo.head.ref.name
 
+  @synchronized
+  def get_branches(self) -> list:
+    return tuple(branch.name for branch in self._repo.branches)
+
+  @synchronized
+  def get_branches_containing(self, rev: str) -> list:
+    try:
+      result = self._repo.git.branch(contains=rev, format="%(refname:short)")
+      return result.splitlines()
+    except git.GitCommandError:
+      return []
+
+  @synchronized
   def get_revisions(self, since_rev: str = None, until_rev: str = "HEAD",  paths: list = None) -> list:
     """
     Returns all revisions having affected {paths} in the ancestry of {until_rev}, bounded by {since_rev} if
@@ -339,6 +448,18 @@ class Repo:
       if commit.parents:
         prev = commit.parents[0]
       diffs = commit.diff(prev, paths=paths, R=True) # R=True -> reverse
-      if len(diffs):
-        revisions.append(Revision(commit, diffs))
+      # Two cases:
+      # - paths were specified: we want the list of commits affecting them (with a diff)
+      # - paths were not: we want the full log
+      if len(diffs) or paths is None:
+        revisions.append(Revision(self, commit, diffs))
     return revisions
+
+  @synchronized
+  def get_tags_pointing_at(self, rev="HEAD") -> list:
+    try:
+      rev = self.rev_parse(rev)
+      result = self._repo.git.tag(points_at=rev)
+      return result.splitlines()
+    except BossmanError:
+      return []
\ No newline at end of file
